@@ -20,7 +20,10 @@
 # along with duplicity; if not, write to the Free Software Foundation,
 # Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 
-from future_builtins import filter, map
+from builtins import next
+from builtins import str
+from builtins import object
+from builtins import filter, map
 
 import os  # @UnusedImport
 import stat  # @UnusedImport
@@ -33,9 +36,9 @@ from duplicity import globals  # @Reimport
 from duplicity import diffdir
 from duplicity import util  # @Reimport
 from duplicity.globmatch import GlobbingError, FilePrefixError, \
-    path_matches_glob_fn
+    select_fn_from_glob
 
-"""Iterate exactly the requested files in a directory
+u"""Iterate exactly the requested files in a directory
 
 Parses includes and excludes to yield correct files.  More
 documentation on what this code does can be found on the man page.
@@ -43,13 +46,8 @@ documentation on what this code does can be found on the man page.
 """
 
 
-class SelectError(Exception):
-    """Some error dealing with the Select class"""
-    pass
-
-
-class Select:
-    """Iterate appropriate Paths in given directory
+class Select(object):
+    u"""Iterate appropriate Paths in given directory
 
     This class acts as an iterator on account of its next() method.
     Basically, it just goes through all the files in a directory in
@@ -83,26 +81,30 @@ class Select:
 
     """
     # This re should not match normal filenames, but usually just globs
-    glob_re = re.compile("(.*[*?[]|ignorecase\\:)", re.I | re.S)
+    glob_re = re.compile(u"(.*[*?[]|ignorecase\\:)", re.I | re.S)
 
     def __init__(self, path):
-        """Initializer, called with Path of root directory"""
+        u"""Initializer, called with Path of root directory"""
         assert isinstance(path, Path), str(path)
         self.selection_functions = []
         self.rootpath = path
-        self.prefix = self.rootpath.name
+        self.prefix = self.rootpath.uc_name
+
+    def __iter__(self):  # pylint: disable=non-iterator-returned
+        return self
+
+    def __next__(self):
+        return next(self.iter)
 
     def set_iter(self):
-        """Initialize generator, prepare to iterate."""
+        u"""Initialize generator, prepare to iterate."""
         # Externally-accessed method
         self.rootpath.setdata()  # this may have changed since Select init
         self.iter = self.Iterate(self.rootpath)
-        self.next = self.iter.next
-        self.__iter__ = lambda: self
         return self
 
     def Iterate(self, path):
-        """Return iterator yielding paths in path
+        u"""Return iterator yielding paths in path
 
         This function looks a bit more complicated than it needs to be
         because it avoids extra recursion (and no extra function calls
@@ -116,20 +118,20 @@ class Select:
             try:
                 mode = os.stat(fullpath)[stat.ST_MODE]
                 if stat.S_ISSOCK(mode):
-                    log.Info(_("Skipping socket %s") % util.ufn(fullpath),
+                    log.Info(_(u"Skipping socket %s") % util.fsdecode(fullpath),
                              log.InfoCode.skipping_socket,
                              util.escape(fullpath))
                 else:
-                    log.Warn(_("Error initializing file %s") % util.ufn(fullpath),
+                    log.Warn(_(u"Error initializing file %s") % util.fsdecode(fullpath),
                              log.WarningCode.cannot_iterate,
                              util.escape(fullpath))
             except OSError:
-                log.Warn(_("Error accessing possibly locked file %s") % util.ufn(fullpath),
+                log.Warn(_(u"Error accessing possibly locked file %s") % util.fsdecode(fullpath),
                          log.WarningCode.cannot_stat, util.escape(fullpath))
             return None
 
         def diryield(path):
-            """Generate relevant files in directory path
+            u"""Generate relevant files in directory path
 
             Returns (path, num) where num == 0 means path should be
             generated normally, num == 1 means the path is a directory
@@ -144,13 +146,13 @@ class Select:
                     error_handler, Path.append, (path, filename))
                 if new_path:
                     s = self.Select(new_path)
-                    if (new_path.type in ["reg", "dir"] and
-                        not os.access(new_path.name, os.R_OK)) and \
-                            (s == 1 or s == 2):
+                    if (new_path.type in [u"reg", u"dir"]
+                        and not os.access(new_path.name, os.R_OK)) \
+                            and (s == 1 or s == 2):
                         # Path is a file or folder that cannot be read, but
                         # should be included or scanned.
-                        log.Warn(_("Error accessing possibly locked file %s") %
-                                 util.ufn(new_path.name),
+                        log.Warn(_(u"Error accessing possibly locked file %s") %
+                                 new_path.uc_name,
                                  log.WarningCode.cannot_read,
                                  util.escape(new_path.name))
                         if diffdir.stats:
@@ -164,10 +166,10 @@ class Select:
 
         if not path.type:
             # base doesn't exist
-            log.Warn(_("Warning: base %s doesn't exist, continuing") %
-                     util.ufn(path.name))
+            log.Warn(_(u"Warning: base %s doesn't exist, continuing") %
+                     path.uc_name)
             return
-        log.Debug(_("Selecting %s") % util.ufn(path.name))
+        log.Debug(_(u"Selecting %s") % path.uc_name)
         yield path
         if not path.isdir():
             return
@@ -176,7 +178,7 @@ class Select:
 
         while diryield_stack:
             try:
-                subpath, val = diryield_stack[-1].next()
+                subpath, val = next(diryield_stack[-1])
             except StopIteration:
                 diryield_stack.pop()
                 if delayed_path_stack:
@@ -185,10 +187,10 @@ class Select:
             if val == 0:
                 if delayed_path_stack:
                     for delayed_path in delayed_path_stack:
-                        log.Log(_("Selecting %s") % util.ufn(delayed_path.name), 6)
+                        log.Log(_(u"Selecting %s") % delayed_path.uc_name, 6)
                         yield delayed_path
                     del delayed_path_stack[:]
-                log.Debug(_("Selecting %s") % util.ufn(subpath.name))
+                log.Debug(_(u"Selecting %s") % subpath.uc_name)
                 yield subpath
                 if subpath.isdir():
                     diryield_stack.append(diryield(subpath))
@@ -197,18 +199,18 @@ class Select:
                 diryield_stack.append(diryield(subpath))
 
     def Select(self, path):
-        """Run through the selection functions and return dominant val 0/1/2"""
+        u"""Run through the selection functions and return dominant val 0/1/2"""
         # Only used by diryield and tests. Internal.
-        log.Debug("Selection: examining path %s" % util.ufn(path.name))
+        log.Debug(u"Selection: examining path %s" % path.uc_name)
         if not self.selection_functions:
-            log.Debug("Selection:     + no selection functions found. Including")
+            log.Debug(u"Selection:     + no selection functions found. Including")
             return 1
         scan_pending = False
         for sf in self.selection_functions:
             result = sf(path)
-            log.Debug("Selection:     result: %4s from function: %s" %
+            log.Debug(u"Selection:     result: %4s from function: %s" %
                       (str(result), sf.name))
-            if result is 2:
+            if result == 2:
                 # Selection function says that the path should be scanned for matching files, but keep going
                 # through the selection functions looking for a real match (0 or 1).
                 scan_pending = True
@@ -224,17 +226,17 @@ class Select:
             result = 1
 
         if result == 0:
-            log.Debug("Selection:     - excluding file")
+            log.Debug(u"Selection:     - excluding file")
         elif result == 1:
-            log.Debug("Selection:     + including file")
+            log.Debug(u"Selection:     + including file")
         else:
             assert result == 2
-            log.Debug("Selection:     ? scanning directory for matches")
+            log.Debug(u"Selection:     ? scanning directory for matches")
 
         return result
 
     def ParseArgs(self, argtuples, filelists):
-        """Create selection functions based on list of tuples
+        u"""Create selection functions based on list of tuples
 
         The tuples are created when the initial commandline arguments
         are read.  They have the form (option string, additional
@@ -246,65 +248,64 @@ class Select:
         filelists_index = 0
         try:
             for opt, arg in argtuples:
-                if opt == "--exclude":
+                if opt == u"--exclude":
                     self.add_selection_func(self.glob_get_sf(arg, 0))
-                elif opt == "--exclude-if-present":
-                    self.add_selection_func(self.present_get_sf(arg, 0))
-                elif opt == "--exclude-device-files":
-                    self.add_selection_func(self.devfiles_get_sf())
-                elif (opt == "--exclude-filelist") or (opt == "--exclude-globbing-filelist"):
+                elif opt == u"--exclude-if-present":
+                    self.add_selection_func(self.present_get_sf(arg, 0), add_to_start=True)
+                elif opt == u"--exclude-device-files":
+                    self.add_selection_func(self.devfiles_get_sf(), add_to_start=True)
+                elif (opt == u"--exclude-filelist") or (opt == u"--exclude-globbing-filelist"):
                     # --exclude-globbing-filelist is now deprecated, as all filelists are globbing
                     # but keep this here for the short term for backwards-compatibility
                     for sf in self.filelist_globbing_get_sfs(filelists[filelists_index], 0, arg):
                         self.add_selection_func(sf)
                     filelists_index += 1
-                elif opt == "--exclude-other-filesystems":
+                elif opt == u"--exclude-other-filesystems":
                     self.add_selection_func(self.other_filesystems_get_sf(0))
-                elif opt == "--exclude-regexp":
+                elif opt == u"--exclude-regexp":
                     self.add_selection_func(self.regexp_get_sf(arg, 0))
-                elif opt == "--exclude-older-than":
+                elif opt == u"--exclude-older-than":
                     self.add_selection_func(self.exclude_older_get_sf(arg))
-                elif opt == "--include":
+                elif opt == u"--include":
                     self.add_selection_func(self.glob_get_sf(arg, 1))
-                elif (opt == "--include-filelist") or (opt == "--include-globbing-filelist"):
+                elif (opt == u"--include-filelist") or (opt == u"--include-globbing-filelist"):
                     # --include-globbing-filelist is now deprecated, as all filelists are globbing
                     # but keep this here for the short term for backwards-compatibility
                     for sf in self.filelist_globbing_get_sfs(filelists[filelists_index], 1, arg):
                         self.add_selection_func(sf)
                     filelists_index += 1
-                elif opt == "--include-regexp":
+                elif opt == u"--include-regexp":
                     self.add_selection_func(self.regexp_get_sf(arg, 1))
                 else:
-                    assert 0, "Bad selection option %s" % opt
-        except SelectError as e:
+                    assert 0, u"Bad selection option %s" % opt
+        except GlobbingError as e:
             self.parse_catch_error(e)
         assert filelists_index == len(filelists)
         self.parse_last_excludes()
 
     def parse_catch_error(self, exc):
-        """Deal with selection error exc"""
+        u"""Deal with selection error exc"""
         # Internal, used by ParseArgs.
         if isinstance(exc, FilePrefixError):
-            log.FatalError(_("""\
+            log.FatalError(_(u"""\
 Fatal Error: The file specification
     %s
 cannot match any files in the base directory
     %s
 Useful file specifications begin with the base directory or some
 pattern (such as '**') which matches the base directory.""") %
-                           (exc, util.ufn(self.prefix)), log.ErrorCode.file_prefix_error)
+                           (exc, self.prefix), log.ErrorCode.file_prefix_error)
         elif isinstance(exc, GlobbingError):
-            log.FatalError(_("Fatal Error while processing expression\n"
-                             "%s") % exc, log.ErrorCode.globbing_error)
+            log.FatalError(_(u"Fatal Error while processing expression\n"
+                             u"%s") % exc, log.ErrorCode.globbing_error)
         else:
             raise  # pylint: disable=misplaced-bare-raise
 
     def parse_last_excludes(self):
-        """Exit with error if last selection function isn't an exclude"""
+        u"""Exit with error if last selection function isn't an exclude"""
         # Internal. Used by ParseArgs.
-        if (self.selection_functions and
-                not self.selection_functions[-1].exclude):
-            log.FatalError(_("""\
+        if (self.selection_functions and not self.selection_functions[-1].exclude):
+            log.FatalError(_(u"""\
 Last selection expression:
     %s
 only specifies that files be included.  Because the default is to
@@ -314,7 +315,7 @@ probably isn't what you meant.""") %
                            log.ErrorCode.redundant_inclusion)
 
     def add_selection_func(self, sel_func, add_to_start=None):
-        """Add another selection function at the end or beginning"""
+        u"""Add another selection function at the end or beginning"""
         # Internal. Used by ParseArgs.
         if add_to_start:
             self.selection_functions.insert(0, sel_func)
@@ -322,7 +323,7 @@ probably isn't what you meant.""") %
             self.selection_functions.append(sel_func)
 
     def filelist_sanitise_line(self, line, include_default):
-        """
+        u"""
         Sanitises lines of both normal and globbing filelists, returning (line, include) and line=None if blank/comment
 
         The aim is to parse filelists in a consistent way, prior to the interpretation of globbing statements.
@@ -333,25 +334,25 @@ probably isn't what you meant.""") %
         line = line.strip()
         if not line:  # skip blanks
             return None, include_default
-        if line[0] == "#":  # skip full-line comments
+        if line.startswith(u"#"):  # skip full-line comments
             return None, include_default
 
         include = include_default
-        if line[:2] == "+ ":
+        if line.startswith(u"+ "):
             # Check for "+ " or "- " syntax
             include = 1
             line = line[2:]
-        elif line[:2] == "- ":
+        elif line.startswith(u"- "):
             include = 0
             line = line[2:]
 
-        if (line[:1] == "'" and line[-1:] == "'") or (line[:1] == '"' and line[-1:] == '"'):
+        if (line.startswith(u"'") and line.endswith(u"'")) or (line.startswith(u'"') and line.endswith(u'"')):
             line = line[1:-1]
 
         return line, include
 
     def filelist_globbing_get_sfs(self, filelist_fp, inc_default, list_name):
-        """Return list of selection functions by reading fileobj
+        u"""Return list of selection functions by reading fileobj
 
         filelist_fp should be an open file object
         inc_default is true iff this is an include list
@@ -360,8 +361,8 @@ probably isn't what you meant.""") %
 
         """
         # Internal. Used by ParseArgs.
-        log.Notice(_("Reading globbing filelist %s") % list_name)
-        separator = globals.null_separator and "\0" or "\n"
+        log.Notice(_(u"Reading globbing filelist %s") % list_name)
+        separator = globals.null_separator and u"\0" or u"\n"
         filelist_fp.seek(0)
         for line in filelist_fp.read().split(separator):
             line, include = self.filelist_sanitise_line(line, inc_default)
@@ -371,7 +372,7 @@ probably isn't what you meant.""") %
             yield self.glob_get_sf(line, include)
 
     def other_filesystems_get_sf(self, include):
-        """Return selection function matching files on other filesystems"""
+        u"""Return selection function matching files on other filesystems"""
         # Internal. Used by ParseArgs and unit tests.
         assert include == 0 or include == 1
         root_devloc = self.rootpath.getdevloc()
@@ -383,36 +384,32 @@ probably isn't what you meant.""") %
                 return None
 
         sel_func.exclude = not include
-        sel_func.name = "Match other filesystems"
+        sel_func.name = u"Match other filesystems"
         return sel_func
 
     def regexp_get_sf(self, regexp_string, include):
-        """Return selection function given by regexp_string"""
+        u"""Return selection function given by regexp_string"""
         # Internal. Used by ParseArgs and unit tests.
         assert include == 0 or include == 1
         try:
             regexp = re.compile(regexp_string)
         except Exception:
-            log.Warn(_("Error compiling regular expression %s") % regexp_string)
+            log.Warn(_(u"Error compiling regular expression %s") % regexp_string)
             raise
 
         def sel_func(path):
-            if regexp.search(path.name):
+            if regexp.search(path.uc_name):
                 return include
             else:
                 return None
 
         sel_func.exclude = not include
-        sel_func.name = "Regular expression: %s" % regexp_string
+        sel_func.name = u"Regular expression: %s" % regexp_string
         return sel_func
 
     def devfiles_get_sf(self):
-        """Return a selection function to exclude all dev files"""
+        u"""Return a selection function to exclude all dev files"""
         # Internal. Used by ParseArgs.
-        if self.selection_functions:
-            log.Warn(_("Warning: exclude-device-files is not the first "
-                       "selector.\nThis may not be what you intended"))
-
         def sel_func(path):
             if path.isdev():
                 return 0
@@ -420,28 +417,26 @@ probably isn't what you meant.""") %
                 return None
 
         sel_func.exclude = 1
-        sel_func.name = "Exclude device files"
+        sel_func.name = u"Exclude device files"
         return sel_func
 
     def glob_get_sf(self, glob_str, include):
-        """Return selection function given by glob string"""
+        u"""Return selection function given by glob string"""
         # Internal. Used by ParseArgs, filelist_globbing_get_sfs and unit tests.
         assert include == 0 or include == 1
-        if glob_str == "**":
+        assert isinstance(glob_str, str)
+        if glob_str == u"**":
             sel_func = lambda path: include
-        elif not self.glob_re.match(glob_str):
-            # normal file
-            sel_func = self.glob_get_filename_sf(glob_str, include)
         else:
             sel_func = self.glob_get_normal_sf(glob_str, include)
 
         sel_func.exclude = not include
-        sel_func.name = "Command-line %s glob: %s" % \
-                        (include and "include" or "exclude", glob_str)
+        sel_func.name = u"Command-line %s glob: %s" % \
+                        (include and u"include" or u"exclude", glob_str)
         return sel_func
 
     def present_get_sf(self, filename, include):
-        """Return selection function given by existence of a file in a directory"""
+        u"""Return selection function given by existence of a file in a directory"""
         # Internal. Used by ParseArgs.
         assert include == 0 or include == 1
 
@@ -456,9 +451,8 @@ probably isn't what you meant.""") %
                     # functions. Currently this will give an error for any
                     # locked directory within the folder being backed up.
                     log.Warn(_(
-                        "Error accessing possibly locked file %s") % util.ufn(
-                        path.name),
-                        log.WarningCode.cannot_read, util.escape(path.name))
+                        u"Error accessing possibly locked file %s") % path.uc_name,
+                        log.WarningCode.cannot_read, util.escape(path.uc_name))
                     if diffdir.stats:
                         diffdir.stats.Errors += 1
                 elif path.append(filename).exists():
@@ -473,72 +467,12 @@ probably isn't what you meant.""") %
                            log.ErrorCode.not_implemented)
 
         sel_func.exclude = not include
-        sel_func.name = "Command-line %s filename: %s" % \
-                        (include and "include-if-present" or "exclude-if-present", filename)
-        return sel_func
-
-    def glob_get_filename_sf(self, filename, include):
-        """Get a selection function given a normal filename
-
-        Some of the parsing is better explained in
-        filelist_parse_line.  The reason this is split from normal
-        globbing is things are a lot less complicated if no special
-        globbing characters are used.
-
-        """
-        # Internal. Used by glob_get_sf and unit tests.
-        # ToDo: Make all globbing/non-globbing use same code path
-        # This distinction has bitten us too many times with bugs in one or
-        # the other.
-        match_only_dirs = False
-
-        if filename != "/" and filename[-1] == "/":
-            match_only_dirs = True
-            # Remove trailing / from directory name (unless that is the entire
-            # string)
-            filename = filename[:-1]
-
-        if not filename.startswith(self.prefix):
-            raise FilePrefixError(filename)
-        index = tuple(filter(lambda x: x,
-                             filename[len(self.prefix):].split("/")))
-        return self.glob_get_tuple_sf(index, include, match_only_dirs)
-
-    def glob_get_tuple_sf(self, tuple, include, match_only_dirs=False):
-        """Return selection function based on tuple"""
-        # Internal. Used by glob_get_filename_sf.
-
-        def include_sel_func(path):
-            if len(tuple) == len(path.index) and match_only_dirs and not path.isdir():
-                # If we are assessing the actual directory (rather than the
-                # contents of the directory) and the glob ended with a /,
-                # only match directories
-                return None
-            elif (path.index == tuple[:len(path.index)] or
-                    path.index[:len(tuple)] == tuple):
-                return 1  # /foo/bar implicitly matches /foo, vice-versa
-            else:
-                return None
-
-        def exclude_sel_func(path):
-            if match_only_dirs and not path.isdir():
-                # If the glob ended with a /, only match directories
-                return None
-            elif path.index[:len(tuple)] == tuple:
-                return 0  # /foo excludes /foo/bar, not vice-versa
-            else:
-                return None
-
-        if include == 1:
-            sel_func = include_sel_func
-        elif include == 0:
-            sel_func = exclude_sel_func
-        sel_func.exclude = not include
-        sel_func.name = "Tuple select %s" % (tuple,)
+        sel_func.name = u"Command-line %s filename: %s" % \
+                        (include and u"include-if-present" or u"exclude-if-present", filename)
         return sel_func
 
     def glob_get_normal_sf(self, glob_str, include):
-        """Return selection function based on glob_str
+        u"""Return selection function based on glob_str
 
         The basic idea is to turn glob_str into a regular expression,
         and just use the normal regular expression.  There is a
@@ -552,24 +486,27 @@ probably isn't what you meant.""") %
         things similar to this.
 
         """
-        # Internal. Used by glob_get_sf and unit tests.
-
+        assert isinstance(glob_str, str), \
+            u"The glob string " + glob_str.decode(sys.getfilesystemencoding(), u"ignore") + u" is not unicode"
         ignore_case = False
 
-        if glob_str.lower().startswith("ignorecase:"):
+        if glob_str.lower().startswith(u"ignorecase:"):
             # Glob string starts with ignorecase, so remove that from the
             # string and change it to lowercase.
-            glob_str = glob_str[len("ignorecase:"):].lower()
+            glob_str = glob_str[len(u"ignorecase:"):].lower()
             ignore_case = True
 
-        # Check to make sure prefix is ok
-        if not path_matches_glob_fn(glob_str, include=1)(self.rootpath):
+        # Check to make sure prefix is ok, i.e. the glob string is within
+        # the root folder being backed up
+        file_prefix_selection = select_fn_from_glob(glob_str, include=1)(self.rootpath)
+        if not file_prefix_selection:
+            # file_prefix_selection == 1 (include) or 2 (scan)
             raise FilePrefixError(glob_str)
 
-        return path_matches_glob_fn(glob_str, include, ignore_case)
+        return select_fn_from_glob(glob_str, include, ignore_case)
 
     def exclude_older_get_sf(self, date):
-        """Return selection function based on files older than modification date """
+        u"""Return selection function based on files older than modification date """
         # Internal. Used by ParseArgs.
 
         def sel_func(path):
@@ -583,5 +520,5 @@ probably isn't what you meant.""") %
             return None
 
         sel_func.exclude = True
-        sel_func.name = "Select older than %s" % (date,)
+        sel_func.name = u"Select older than %s" % (date,)
         return sel_func
